/*

    Copyright (c) 2018 - 2020 Analog Flavor EURL

    Contact us for support:
        http://www.analogflavor.com

-----------------------------------------------------------------------------
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    * Neither the name of Analog Flavor EURL nor 
      the names of its contributors may be used to endorse or promote products derived from
      this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
-----------------------------------------------------------------------------

*/


#include <wiener_mpod_lv_hv_snmp.h>

#include <string.h>
#include <math.h>
#include <stdio.h>

#ifdef WIN32
#include <Windows.h>
void sleep_ms(int ms) {
    Sleep(ms);
}
#else
#include <unistd.h>
void sleep_ms(int ms) {
    usleep(ms * 1000);
}
#endif

int main(int argc, char **argv) {

    void *obj_ptr;
    const char *string_value;
    int int_value;
    const int *occupied_slots;
    const int *occupied_slots_iterator;
    unsigned module_slot_index, nb_channels_per_module;
    int thread_polling_interval_ms = 1000; /* set to 0 for debug */

    if (!mpod_snmp_winsock_startup()) {
        printf("Winsock startup failed.\n");
        getchar();
        return 1;
    }

    /* create object */
    obj_ptr = mpod_snmp_create();

    /*  establish connection */
    /* "guru" is the default snmp community */
    if (argc >= 2) {
        /* the port ans community can also specified as  "192.168.0.101:161:guru" only in the first argument. */
        mpod_snmp_crateConnect(obj_ptr, argv[1], 0, "", 0.5 * thread_polling_interval_ms, thread_polling_interval_ms, 2.0 * thread_polling_interval_ms);
    }
    else {
        mpod_snmp_crateConnect(obj_ptr, "192.168.0.101", 161, "", 0.5 * thread_polling_interval_ms, thread_polling_interval_ms, 2.0 * thread_polling_interval_ms);
    }

    /* check that the connection has been established */
    if (mpod_snmp_crateCheckConnectionIsOk(obj_ptr)) {

        /* make sure all data has been read to the buffer */
        sleep_ms(3 * thread_polling_interval_ms);

        /* get the list of slots occupied by a LV or HV module */
        /* the array is terminated by a -1 */
        occupied_slots = mpod_snmp_crateGetOccupiedSlots(obj_ptr);
        if (occupied_slots) {
            /* print the list of occupied slots */
            occupied_slots_iterator = occupied_slots;
            printf("CrateGetOccupiedSlots() returns [ ");
            while (*occupied_slots_iterator >= 0) {
                printf("%d ", *occupied_slots_iterator);
                occupied_slots_iterator++;
            }
            printf("]\n");
            /* traverse the list of occupied slots */
            occupied_slots_iterator = occupied_slots;
            while (*occupied_slots_iterator >= 0) {
                module_slot_index = *occupied_slots_iterator;
                /* find out how many channels are available for each module */
                nb_channels_per_module = mpod_snmp_moduleGetNumberOfChannels(obj_ptr, module_slot_index);
                printf("ModuleGetNumberOfChannels(%d) returns %d.\n", module_slot_index, nb_channels_per_module);
                /* find out if the module is a HV module or not */
                int_value = mpod_snmp_moduleGetIsHvModule(obj_ptr, module_slot_index);
                printf("ModuleGetIsHvModule(%d) returns %d.\n", module_slot_index, int_value);
                occupied_slots_iterator++;
            }
        }
        else {
            printf("CrateGetOccupiedSlots() fails.\n");
        }

    }
    else {

        /* print error messages */
        printf("CrateCheckConnectionIsOk() returns false.\n");
        int_value = mpod_snmp_crateGetConnectionStatus(obj_ptr);
        printf("CrateGetConnectionStatus() returns %d.\n", int_value);
        string_value = mpod_snmp_crateGetConnectionStatusAsString(obj_ptr);
        printf("CrateGetConnectionStatusAsString() returns %s.\n", string_value);
        string_value = mpod_snmp_crateGetErrorString(obj_ptr);
        printf("CrateGetErrorString() returns %s.\n", string_value);

    }

    /* delete object */
    mpod_snmp_delete(obj_ptr);

    mpod_snmp_winsock_cleanup();

#ifdef WIN32
    getchar();
#endif

    return 0;
}
