#
#    Copyright (c) 2018 - 2020 Analog Flavor EURL
#
#    Contact us for support:
#        http://www.analogflavor.com
#
# -----------------------------------------------------------------------------
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#    * Neither the name of Analog Flavor EURL nor 
#      the names of its contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------
#


## \file 0002-low-level-snmp.py
#  \brief This file contains an example for the use of the WienerMpodLvHvSnmpPython module.
#
#  The used functions allow to control and monitor a single Wiener MPOD crate over SNMP.
#
#  The comments in the example code give more explanations on the performed API function calls.

import sys
import time
import math
import argparse
import import_mpod_snmp_module

thread_polling_interval_ms = 1000.0
wait_delay = 3 * thread_polling_interval_ms * 0.001

# Here we build the relative path to the python module.
# The path might have to be adjusted if the example is run outside the 
# Analog Flavor software installation tree.
mpod_snmp = import_mpod_snmp_module.import_mpod_snmp_module()

# the IP address, port and snmp community can be specified in the first argument as
#           crateConnect("192.168.0.101:161:guru", ...
#       or in the first 3 arguments as
#           crateConnect("192.168.0.101", 161, "guru", ...
# the polling intervals are optional arguments
if (len(sys.argv) >= 2) :
    mpod_crate = mpod_snmp.crateConnect(sys.argv[1], 0, "", 0.5 * thread_polling_interval_ms, thread_polling_interval_ms, 2.0 * thread_polling_interval_ms);
else :
    print("Please specify an MPOD connection on the comnmand line (ex. \"192.168.0.100:161:guru\").")
    default_crate_ip = import_mpod_snmp_module.get_default_crate_ip()
    mpod_crate = mpod_snmp.crateConnect(default_crate_ip, 0, "")

print("newly created crate  == {0}".format(mpod_crate));

# check that the connection has been established 
if (mpod_crate.checkConnectionIsOk()) :

    # wait to make sure all data has been read to the buffer 
    time.sleep(wait_delay)

    # command: snmpGetDescription
    #     input: snmp timeout interval in ms or -1 to use the default value
    #     output: True for success.
    result = mpod_crate.snmpSetTimeoutInterval(5000)
    print("snmpSetTimeoutInterval() returned \"{0}\"".format(result))

    # command: snmpGetDescription
    #     input: Oid as string
    #     output: Description from MIB file.
    Oid = "iso.3.6.1.4.1.19947.1.3.2.1.2.101"
    result = mpod_crate.snmpGetDescription(Oid)
    print("snmpGetDescription({0}) returned \"{1}\"".format(Oid, result))
    Oid = "WIENER-CRATE-MIB::outputName.u100"
    result = mpod_crate.snmpGetDescription(Oid)
    print("snmpGetDescription({0}) returned \"{1}\"".format(Oid, result))

    # command: snmpGetNextOid
    #     input: CurrentOid
    #     output: NextOid
    Oid = 'iso'
    result = mpod_crate.snmpGetNextOid(Oid)
    print("snmpGetNextOid({0}) returned \"{1}\"".format(Oid, result))
    Oid = 'WIENER-CRATE-MIB::moduleAuxiliaryMeasurementVoltage0'
    result = mpod_crate.snmpGetNextOid(Oid)
    print("snmpGetNextOid({0}) returned \"{1}\"".format(Oid, result))

    # command: SnmpGetValueDouble
    #     input: Oid name
    #     output: Readout
    Oid = 'iso.3.6.1.4.1.19947.1.3.2.1.5.101'
    result = mpod_crate.snmpGetValueDouble(Oid)
    print("snmpGetValueDouble({0}) returned \"{1}\"".format(Oid, result))
    Oid = 'WIENER-CRATE-MIB::outputMeasurementSenseVoltage.u100'
    result = mpod_crate.snmpGetValueDouble(Oid)
    print("snmpGetValueDouble({0}) returned \"{1}\"".format(Oid, result))

    # command: SnmpGetValueInt
    #     input: Oid name
    #     output: Readout
    Oid = 'iso.3.6.1.4.1.19947.1.1.1.0'
    result = mpod_crate.snmpGetValueInt(Oid)
    print("snmpGetValueInt({0}) returned \"{1}\"".format(Oid, result))
    Oid = 'WIENER-CRATE-MIB::sysMainSwitch.0'
    result = mpod_crate.snmpGetValueInt(Oid)
    print("snmpGetValueInt({0}) returned \"{1}\"".format(Oid, result))

    # command: SnmpGetValueString
    #     input: Oid name
    #     output: Readout
    Oid = 'iso.3.6.1.2.1.1.1.0'
    result = mpod_crate.snmpGetValueString(Oid)
    print("snmpGetValueInt({0}) returned \"{1}\"".format(Oid, result))
    Oid = 'SNMPv2-SMI::mib-2.1.1.0'
    result = mpod_crate.snmpGetValueString(Oid)
    print("snmpGetValueInt({0}) returned \"{1}\"".format(Oid, result))

    # command: SnmpSetValueDouble
    #     input: Variable name and value to set
    #     output: Success
    Oid = 'iso.3.6.1.4.1.19947.1.3.2.1.10.301'
    result = mpod_crate.snmpSetValueDouble(Oid, 1027.0)
    print("snmpSetValueDouble({0}) returned \"{1}\"".format(Oid, result))
    ## it takes some time until the the new value is taken into account (write and read to device)
    time.sleep(wait_delay)
    result = mpod_crate.snmpGetValueDouble(Oid)
    print("snmpGetValueDouble({0}) returned \"{1}\"".format(Oid, result))

    Oid = 'WIENER-CRATE-MIB::outputVoltage.u300'
    result = mpod_crate.snmpSetValueDouble(Oid, 527.0)
    print("snmpSetValueDouble({0}) returned \"{1}\"".format(Oid, result))
    ## it takes some time until the the new value is taken into account (write and read to device)
    time.sleep(wait_delay)
    result = mpod_crate.snmpGetValueDouble(Oid)
    print("snmpGetValueDouble({0}) returned \"{1}\"".format(Oid, result))

    # command: SnmpSetValueInt
    #     input: Variable name and value to set
    #     output: Success
    Oid = 'iso.3.6.1.4.1.19947.1.1.1.0'
    result = mpod_crate.snmpSetValueInt(Oid, 0)
    ## it takes some time until the the new value is taken into account (write and read to device)
    time.sleep(wait_delay)
    result = mpod_crate.snmpGetValueInt(Oid)
    print("snmpGetValueInt({0}) returned \"{1}\"".format(Oid, result))

    Oid = 'WIENER-CRATE-MIB::sysMainSwitch.0'
    result = mpod_crate.snmpSetValueInt(Oid, 0)
    ## it takes some time until the the new value is taken into account (write and read to device)
    time.sleep(wait_delay)
    result = mpod_crate.snmpGetValueInt(Oid)
    print("snmpGetValueInt({0}) returned \"{1}\"".format(Oid, result))

    # command: SnmpTranslateFromOid
    #     input: Oid as string
    #     output: Human readable string
    Oid = 'iso.3.6.1.4.1.19947.1.1.1.0'
    result = mpod_crate.snmpTranslateFromOid(Oid)
    print("snmpTranslateFromOid({0}) returned \"{1}\"".format(Oid, result))
    Oid = 'WIENER-CRATE-MIB::sysMainSwitch.0'
    result = mpod_crate.snmpTranslateFromOid(Oid)
    print("snmpTranslateFromOid({0}) returned \"{1}\"".format(Oid, result))

    # command: SnmpTranslateToOid
    #     input: Human readable string
    #     output: Oid as string
    Oid = 'iso.3.6.1.4.1.19947.1.1.1.0'
    result = mpod_crate.snmpTranslateToOid(Oid)
    print("snmpTranslateToOid({0}) returned \"{1}\"".format(Oid, result))
    Oid = 'WIENER-CRATE-MIB::sysMainSwitch.0'
    result = mpod_crate.snmpTranslateToOid(Oid)
    print("snmpTranslateToOid({0}) returned \"{1}\"".format(Oid, result))

else :
    # print error messages 
    print("checkConnectionIsOk() returns false.");
    int_value = mpod_crate.getConnectionStatus()
    print("getConnectionStatus() returns {0}.".format(int_value))
    string_value = mpod_crate.getConnectionStatusAsString()
    print("getConnectionStatusAsString() returns {0}.".format(string_value))
    string_value = mpod_crate.getErrorString()
    print("getErrorString() returns {0}.".format(string_value))

exit(0)

