#
#    Copyright (c) 2018 - 2020 Analog Flavor EURL
#
#    Contact us for support:
#        http://www.analogflavor.com
#
# -----------------------------------------------------------------------------
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#    * Neither the name of Analog Flavor EURL nor 
#      the names of its contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------
#


## \file 0003-high-level-crate-functions.py
#  \brief This file contains an example for the use of the WienerMpodLvHvSnmpPython module.
#
#  The used functions allow to control and monitor a single Wiener MPOD crate over SNMP.
#
#  The comments in the example code give more explanations on the performed API function calls.

import sys
import time
import math
import argparse
import import_mpod_snmp_module

thread_polling_interval_ms = 1000.0
wait_delay = 3 * thread_polling_interval_ms * 0.001

# Here we build the relative path to the python module.
# The path might have to be adjusted if the example is run outside the 
# Analog Flavor software installation tree.
mpod_snmp = import_mpod_snmp_module.import_mpod_snmp_module()

# the IP address, port and snmp community can be specified in the first argument as
#           crateConnect("192.168.0.101:161:guru", ...
#       or in the first 3 arguments as
#           crateConnect("192.168.0.101", 161, "guru", ...
# the polling intervals are optional arguments
if (len(sys.argv) >= 2) :
    mpod_crate = mpod_snmp.crateConnect(sys.argv[1], 0, "", 0.5 * thread_polling_interval_ms, thread_polling_interval_ms, 2.0 * thread_polling_interval_ms);
else :
    print("Please specify an MPOD connection on the comnmand line (ex. \"192.168.0.100:161:guru\").")
    default_crate_ip = import_mpod_snmp_module.get_default_crate_ip()
    mpod_crate = mpod_snmp.crateConnect(default_crate_ip, 0, "")

print("newly created crate  == {0}".format(mpod_crate))

# check that the connection has been established 
if (mpod_crate.checkConnectionIsOk()) :

    # wait to make sure all data has been read to the buffer 
    time.sleep(wait_delay)

    # command: getConnectionStatus
    #     input: -
    #     output: integer value
    result = mpod_crate.getConnectionStatus()
    print("getConnectionStatus() returns \"{0}\"".format(result))

    # command: getConnectionStatusAsString
    #     input: -
    #     output: string value
    result = mpod_crate.getConnectionStatusAsString()
    print("getConnectionStatusAsString() returns \"{0}\"".format(result))

    # command: checkConnectionIsOk
    #     input: -
    #     output: True of False
    result = mpod_crate.checkConnectionIsOk()
    print("checkConnectionIsOk() returns \"{0}\"".format(result))

    # command: getErrorString
    #     input: -
    #     output: string value
    result = mpod_crate.getErrorString()
    print("getErrorString() returns \"{0}\"".format(result))

    # command: checkIsMibFileLoaded
    #     input: -
    #     output: True or False
    result = mpod_crate.checkIsMibFileLoaded()
    print("checkIsMibFileLoaded() returns \"{0}\"".format(result))

    # command: getNumberOfOccupiedSlots
    #     input: -
    #     output: integer value
    result = mpod_crate.getNumberOfOccupiedSlots()
    print("getNumberOfOccupiedSlots() returns \"{0}\"".format(result))

    # command: getOccupiedSlots
    #     input: -
    #     output: list of integer values
    result = mpod_crate.getOccupiedSlots()
    print("getOccupiedSlots() returns \"{0}\"".format(result))

    # command: getSysDescr
    #     input: -
    #     output: string value
    result = mpod_crate.getSysDescr()
    print("getSysDescr() returns \"{0}\"".format(result))

    # command: groupGetIndex
    #     input: integer value
    #     output: integer value
    for i in range(0, 256) :
        result = mpod_crate.groupGetIndex(i)
        if result != -1 :
            print("groupGetIndex({0}) returns \"{1}\"".format(i, result))

    # command: groupSetSwitch
    #     input: group,_index, value
    #     output: integer value
    result = mpod_crate.groupSetSwitch(0, 0)
    print("groupSetSwitch(0, 0) returns \"{0}\"".format(result))

    # command: getFanSpeed
    #     input: -
    #     output: integer value
    result = mpod_crate.getFanSpeed()
    print("getFanSpeed() returns \"{0}\"".format(result))

    # command: setFanSpeed
    #     input: Value
    #     output: True for success
    new_fan_speed = 1500
    result = mpod_crate.setFanSpeed(new_fan_speed)
    print("setFanSpeed({0}) returns \"{1}\"".format(new_fan_speed, result))
    result = mpod_crate.getFanSpeed()
    print("getFanSpeed() returns \"{0}\"".format(result))

    # command: getNumberOfChannels
    #     input: -
    #     output: integer
    # (remark) this returns the total number of channels (sum of all module channels).
    result = mpod_crate.getNumberOfChannels()
    print("getNumberOfChannels() returns \"{0}\"".format(result))

    # command: getNumberOfGroups
    #     input: -
    #     output: integer
    result = mpod_crate.getNumberOfGroups()
    print("getNumberOfGroups() returns \"{0}\"".format(result))

    # command: getNumberOfModules
    #     input: -
    #     output: integer
    result = mpod_crate.getNumberOfModules()
    print("getNumberOfModules() returns \"{0}\"".format(result))

    # command: getSerialNumber
    #     input: -
    #     output: string
    result = mpod_crate.getSerialNumber()
    print("getSerialNumber() returns \"{0}\"".format(result))

    # command: getSysMainSwitch
    #     input: -
    #     output: integer
    result = mpod_crate.getSysMainSwitch()
    print("getSysMainSwitch() returns \"{0}\"".format(result))

    # command: getSysStatus
    #     input: -
    #     output: integer
    result = mpod_crate.getSysStatus()
    print("getSysStatus() returns \"{0}\"".format(result))

    # command: getSysStatusAsString
    #     input: -
    #     output: string
    result = mpod_crate.getSysStatusAsString()
    print("getSysStatusAsString() returns \"{0}\"".format(result))

    # command: setSysHardwareReset
    #     input: Value
    #     output: True if successfull
    # result = mpod_crate.setSysHardwareReset()
    # print("setSysHardwareReset() returns \"{0}\"".format(result))

    # command: setSysMainSwitch
    #     input: Value
    #     output: True if successfull
    result = mpod_crate.setSysMainSwitch(1)
    print("setSysMainSwitch(1) returns \"{0}\"".format(result))
    ## it takes some time until the the new value is taken into account (write and read to device)
    print(" ... waiting before comparing read and written value ...")
    time.sleep(wait_delay)
    result = mpod_crate.getSysMainSwitch()
    print("getSysMainSwitch() returns \"{0}\"".format(result))

    # command: disconnect
    #     input: -
    #     output: True if successfull
    result = mpod_crate.disconnect()
    print("disconnect() returns \"{0}\"".format(result))

else :
    # print error messages 
    print("checkConnectionIsOk() returns false.");
    int_value = mpod_crate.getConnectionStatus()
    print("getConnectionStatus() returns {0}.".format(int_value))
    string_value = mpod_crate.getConnectionStatusAsString()
    print("getConnectionStatusAsString() returns {0}.".format(string_value))
    string_value = mpod_crate.getErrorString()
    print("getErrorString() returns {0}.".format(string_value))

exit(0)


